#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from irpf90_t import *
from util import *
from variables import variables
from modules import modules

FILENAME=irpdir+'irp_touches.irp.F90'

def create():
  out = []
  l = variables.keys()
  l.sort
  main_modules = filter(lambda x: modules[x].is_main, modules)
  finalize = "subroutine irp_finalize_%s\n"%(irp_id)
  for m in filter(lambda x: not modules[x].is_main, modules):
    finalize += " use %s\n"%(modules[m].name)
  for v in l:
    var = variables[v]
    var_in_main = False
    for m in main_modules:
      if var.fmodule == modules[m].name:
        var_in_main = True
        break
    if not var_in_main:
      if var.is_touched:
        out += var.toucher
      if var.dim != []:
        finalize += "  if (allocated(%s)) then\n"%v
        finalize += "    %s_is_built = .False.\n"%var.same_as
        finalize += "    deallocate(%s)\n"%v
        finalize += "  endif\n"
  finalize += "end\n"


  if out != []:
    out = map(lambda x: "%s\n"%(x),out)

  out += finalize
  
  if not same_file(FILENAME,out):
    file = open(FILENAME,'w')
    file.writelines(out)
    file.close()

if __name__ == '__main__':
  create()




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import os

to_remove = """__init__.py cython_setup.py version.py command_line.py""".split()
ext_modules = []

files = os.listdir('.')
for file in to_remove:
  files.remove(file)

for file in files:
  if file.endswith(".py"):
   module = file.split('.')[0]
   ext_modules += [ Extension(module,[file]) ]

setup(
  name = 'IRPF90 extensions',
  cmdclass = {'build_ext': build_ext},
  ext_modules = ext_modules
)







#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from variable import *
from irpf90_t import *
from command_line import command_line
from util import *

######################################################################
def create_variables():
  from preprocessed_text import preprocessed_text
  result = {}
  icount = 0
  for filename, text in preprocessed_text:
    buffer = []
    inside = False
    for line in text:
      if type(line) == Begin_provider:
        inside = True
      if inside:
        buffer.append(line)
      if type(line) == End_provider:
        inside = False
        icount += 1
        v = Variable(buffer,icount)
        if v.name in result:
          print "Warning: Duplicate provider for %s in"%(v.name)
          print "- ", v.line.filename[0], " line ", v.line.i
          print "- ", result[v.name].line.filename[0], " line ", result[v.name].line.i
          print "Choosing first version"
        result[v.name] = v
        for other in v.others:
          if other in result:
            print "Warning: Duplicate provider for %s in"%(other)
            print "- ", v.line.filename[0], " line ", v.line.i
            print "- ", result[other].line.filename[0], " line ", result[other].line.i
            print "Choosing first version"
          result[other] = Variable(buffer,icount,other)
        buffer = []
  return result

variables = create_variables()

######################################################################
def build_use(vars):
  result = map(lambda x: "  use %s"%(variables[x].fmodule), vars)
  result = make_single(result)
  return result

######################################################################
def call_provides(vars,opt=False):
  vars = make_single( map(lambda x: variables[x].same_as, vars) )
  if opt:
    all_children = flatten( map(lambda x: variables[x].children, vars ))
    vars = filter(lambda x: x not in all_children,vars)
  def fun(x):
    result = []
    result += [ \
    "  if (.not.%s_is_built) then"%(x) ]
    result += [ \
    "    call provide_%s"%(x) ]
    result += [ \
    "  endif" ]
    return result

  result = flatten ( map (fun, vars) )
  return result

######################################################################
if __name__ == '__main__':
  for v in variables.keys():
    print v



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr



import vim
import os,sys
try:
  wd = os.path.abspath(os.path.dirname(__file__))
  sys.path.insert(0,(wd+"/../src/"))
except:
  pass
sys.setcheckinterval(1000)

from command_line import command_line

def main():

  vim.install()

  if command_line.do_help:
    command_line.usage()

  if command_line.do_version:
    from version import version
    print version

  from init import init
  if command_line.do_init:
    init()

  if command_line.do_preprocess:
    init()
    from preprocessed_text import preprocessed_text
    for filename,text in preprocessed_text:
      if filename in command_line.preprocessed:
        for line in text:
          print line.text


  if command_line.do_touch:
    from variables import variables
    for var in command_line.touched:
      if var not in variables:
        print "%s is not an IRP entity"%(var,)
      else:
        print "Touching %s invalidates the following entities:"%(var,)
        parents = variables[var].parents
        parents.sort()
        for x in parents:
          print "- %s"%(x,)


  if not command_line.do_run:
     return

  init()

  import irp_stack
  irp_stack.create()

  import makefile
  makefile.create()

  if command_line.do_codelet:
     import profile
     profile.build_rdtsc()
     import codelet
     codelet.run()

  from modules import modules, write_module
  for m in modules.keys():
    write_module(modules[m])

  makefile.run()

  import touches
  touches.create()

#  import checkpoint
#  checkpoint.create()

  import create_man
  create_man.run()

  if command_line.do_profile:
    import profile
    profile.run()

  if command_line.do_openmp:
    import locks
    locks.create()

if __name__ == '__main__':
  main()




#/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import os

def install():
  VIM = os.environ["HOME"]+"/.vim"
  try:
    if os.access(VIM+"/syntax/irpf90.vim",os.F_OK):
       return
    if not os.access(VIM,os.F_OK):
      os.mkdir(VIM)
    file = open(VIM+"/filetype.vim","a")
    file.write("au BufRead,BufNewFile *.irp.f setfiletype irpf90")
    file.close()
    if not os.access(VIM+"/syntax",os.F_OK):
      os.mkdir(VIM+"/syntax")
    wd = os.path.abspath(os.path.dirname(__file__))
    with open(wd+"/irpf90.vim",'r') as f_in:
        with open(VIM+"/syntax/irpf90.vim",'w') as f_out:
            f_out.write( f_in.read() )
  except:
    pass

if __name__ == "__main__":
  install()



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


from irpf90_t import *
from parsed_text import parsed_text
from module import Fmodule
from util import *

######################################################################
def create_modules():
  result = {}
  for filename,text in parsed_text:
    result[filename] = Fmodule(text,filename)  
  return result

modules = create_modules()

######################################################################
def write_module(m):
  # Module data
  filename = irpdir+m.filename+".irp.module.F90"
  text = m.header + m.head 
  text = map(lambda x: "%s\n"%(x),text)
  if not same_file(filename,text):
#    print filename
    file = open(filename,"w")
    file.writelines(text)
    file.close()

  # Subroutines
  filename = irpdir+m.filename+".irp.F90"
  text = m.header + m.generated_text + m.residual_text
  text = map(lambda x: "%s\n"%(x),text)
  if not same_file(filename,text):
#    print filename
    file = open(filename,"w")
    file.writelines(text)
    file.close()

######################################################################
if __name__ == '__main__':
  write_module(modules['psi.irp.f'])
  




#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import util 
from command_line import command_line

do_debug = command_line.do_debug
do_openmp = command_line.do_openmp
do_memory = command_line.do_memory

import irpf90_t

FILENAME = irpf90_t.irpdir+"irp_stack.irp.F90"

def create():

  txt = """
module irp_stack_mod
  integer, parameter            :: STACKMAX=1000
  character*(128),allocatable   :: irp_stack(:,:)
  double precision,allocatable  :: irp_cpu(:,:)
  integer,allocatable           :: stack_index(:)
  logical                       :: alloc = .False.
  integer                       :: nthread
  character*(128)               :: white = ''
end module

subroutine irp_enter(irp_where)
 use irp_stack_mod
 integer       :: ithread
 character*(*) :: irp_where
"""
  if not do_openmp:
    txt += """
   ithread = 0
"""
  else:
    txt += """
 integer, external :: omp_get_thread_num
 integer, external :: omp_get_num_threads
 ithread = omp_get_thread_num()
"""

  txt += "$1"

  if do_memory:
     txt+="""
 if (.not.alloc) then
"""
     if do_openmp:
       txt += """
 !$OMP PARALLEL
 !$OMP SINGLE
 nthread = omp_get_num_threads()
 !$OMP END SINGLE
 !$OMP END PARALLEL
"""
     else:
       txt += """
 nthread = 1
 """
     txt += """
   print *, 'Allocating irp_stack(',STACKMAX,',',0:nthread,')'
   print *, 'Allocating irp_cpu(',STACKMAX,',',0:nthread,')'
   print *, 'Allocating stack_index(',0:nthread,')'
 endif"""
  txt +="""
$2
end subroutine

subroutine irp_enter_f(irp_where)
 use irp_stack_mod
 integer       :: ithread
 character*(*) :: irp_where
 """
  if do_openmp:
    txt += """
 integer, external :: omp_get_thread_num
 integer, external :: omp_get_num_threads
 ithread = omp_get_thread_num()
"""
  else:
    txt += """
  ithread = 0
"""
  txt += """
$1
"""
  if do_memory:
    txt+="""
 if (.not.alloc) then
"""
    if do_openmp:
      txt += """
 !$OMP PARALLEL
 !$OMP SINGLE
  nthread = omp_get_num_threads()
 !$OMP END SINGLE
 !$OMP END PARALLEL
"""
    else:
      txt += """
  nthread = 1
"""
    txt +="""
  print *, 'Allocating irp_stack(',STACKMAX,',',0:nthread,')'
  print *, 'Allocating irp_cpu(',STACKMAX,',',0:nthread,')'
  print *, 'Allocating stack_index(',0:nthread,')'
 endif
"""
  txt += """
$2
end subroutine

subroutine irp_leave (irp_where)
 use irp_stack_mod
  character*(*) :: irp_where
  integer       :: ithread
  double precision :: cpu
"""
  if do_openmp:
    txt += """
 integer, external :: omp_get_thread_num
 ithread = omp_get_thread_num()
 """
  else:
    txt += """
 ithread = 0
 """
  txt += """
$3
$4
end subroutine
"""

  # $1
  if do_debug:
    s = """
 if (.not.alloc) then
 """
    if do_openmp:
      s += """
 !$OMP PARALLEL
 !$OMP SINGLE
 nthread = omp_get_num_threads()
 !$OMP END SINGLE
 !$OMP END PARALLEL
 !$OMP CRITICAL
 if (.not.alloc) then
   allocate(irp_stack(0:STACKMAX,0:nthread))
   allocate(irp_cpu(0:STACKMAX,0:nthread))
   allocate(stack_index(0:nthread))
   stack_index = 0
   alloc = .True.
 endif
 !$OMP END CRITICAL
 endif
 stack_index(ithread) = min(stack_index(ithread)+1,STACKMAX)
 irp_stack(stack_index(ithread),ithread) = irp_where"""
    else:
      s += """
 nthread = 1
 if (.not.alloc) then
   allocate(irp_stack(0:STACKMAX,1))
   allocate(irp_cpu(0:STACKMAX,1))
   allocate(stack_index(2))
   stack_index = 0
   alloc = .True.
 endif
 endif
 stack_index(1) = min(stack_index(1)+1,STACKMAX)
 irp_stack(stack_index(1),1) = irp_where"""
    if do_memory:
      txt+="""
  print *, 'Allocating irp_stack(',STACKMAX,','0:nthread,')'
  print *, 'Allocating irp_cpu(',STACKMAX,','0:nthread,')'
  print *, 'Allocating stack_index(',0:nthread,')'"""
  else:
    s = ""
  txt = txt.replace("$1",s)

  # $2
  if do_debug:
    txt = txt.replace("$2","""
  print *, ithread, ':', white(1:stack_index(ithread))//'-> ', trim(irp_where)
  call cpu_time(irp_cpu(stack_index(ithread),ithread))""")
  else:
    txt = txt.replace("$2","")

  # $3
  if do_debug:
    txt = txt.replace("$3","""
  call cpu_time(cpu)
  print *, ithread, ':', white(1:stack_index(ithread))//'<- ', &
    trim(irp_stack(stack_index(ithread),ithread)), &
    cpu-irp_cpu(stack_index(ithread),ithread)""")
  else:
    txt = txt.replace("$3","")

  # $4
  if do_debug:
    txt = txt.replace("$4","""
  stack_index(ithread) = max(0,stack_index(ithread)-1)""")
  else:
    txt = txt.replace("$4","")

  txt += """
subroutine irp_trace
 use irp_stack_mod
 integer :: ithread
 integer :: i
"""
  if do_openmp:
    txt += """
!$ integer, external :: omp_get_thread_num
!$ ithread = omp_get_thread_num()
"""
  else:
    txt += """
 ithread = 0
"""
  txt += """
 if (.not.alloc) return
 print *, 'Stack trace: ', ithread
 print *, '-------------------------'
 do i=1,stack_index(ithread)
  print *, trim(irp_stack(i,ithread))
 enddo
 print *, '-------------------------'
end subroutine
"""

  txt = txt.split('\n')
  txt = map(lambda x: x+"\n",txt)
  if not util.same_file(FILENAME, txt):
    file = open(FILENAME,'w')
    file.writelines(txt)
    file.close()





#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne
#   31062 Toulouse Cedex 4
#   scemama@irsamc.ups-tlse.fr


from util import *
from irpf90_t import *
from variables import variables
from preprocessed_text import preprocessed_text
from subroutines import subroutines
import regexps, re
import error

vtuple = map(lambda v: (v, variables[v].same_as, variables[v].regexp), variables.keys())
stuple = map(lambda s: (s, subroutines[s].regexp), subroutines.keys())
stuple = filter(lambda s: subroutines[s[0]].is_function, stuple)
re_string_sub = regexps.re_string.sub

regexps_re_string_sub = regexps.re_string.sub
def find_variables_in_line(line):
  assert isinstance(line,Line)
  result = []
  sub_done = False
  buffer = line.lower
  ap = result.append
  for v,same_as,regexp in vtuple:
    if v in buffer:
      if not sub_done:
        buffer = regexps_re_string_sub('',buffer)
        sub_done = True
      if regexp.search(buffer) is not None:
        ap(same_as)
  return result

def find_funcs_in_line(line):
  assert isinstance(line,Line)
  result = []
  append = result.append
  sub_done = False
  buffer = line.lower
  for s,regexp in stuple:
     if s in buffer:
      if regexp.search(buffer) is not None:
        append(s)
  return result


def find_subroutine_in_line(line):
  assert type(line) == Call
  buffer = line.text.split('(')[0]
  buffer = buffer.split()[1].lower()
  return buffer

def check_touch(line,vars,main_vars):
  def fun(main_var):
    if main_var not in variables:
      error.fail(line,"Variable %s unknown"%(main_var,))
    x = variables[main_var]
    return [main_var]+x.others
  all_others = make_single(flatten( map(fun,main_vars) ))
  all_others.sort()
  vars.sort()
  for x,y in zip(vars,all_others):
      if x != y:
        message = "The following entities should be touched:\n"
        message = "\n".join([message]+map(lambda x: "- %s"%(x,),all_others))
        error.fail(line,message)

################################################################################
def update_variables():
  for filename,text in preprocessed_text:

    for line in filter(lambda x: type(x) in [ Touch, SoftTouch ], text):
        vars = line.lower.split()
        if len(vars) < 2:
          error.fail(line,"Syntax error")
        for v in vars[1:]:
          if v not in variables:
            error.fail(line,"Variable %s unknown"%(v,))
          variables[v]._is_touched = True

    for line in filter(lambda x: type(x) == Free,text):
        vars = line.lower.split()
        if len(vars) < 2:
          error.fail(line,"Syntax error")
        for v in vars[1:]:
          if v not in variables:
            error.fail(line,"Variable %s unknown"%(v,))
          variables[v].is_freed = True

    for line in filter(lambda x: type(x) == Irp_read,text):
        variables[line.filename]._is_read = True

    for line in filter(lambda x: type (x) == Irp_write,text):
        variables[line.filename]._is_written = True

################################################################################

def get_parsed_text():
  def func(filename, text):
    varlist = []
    result = []
    append = result.append
    for line in text: #filter(
#     lambda x: type(x) not in [ Doc, Begin_doc, End_doc ],
#     text):
      if type(line) in [ \
        Empty_line,
        Continue,
        Return,
        Begin_shell,
        End_shell,
        Openmp,
        Directive,
        Use,
        Enddo,
        End_select,
        Endif,
        Implicit,
        Program,
        Subroutine,
        Function,
        End,
      ]:
        append( ([],line) )
      elif type(line) in [ Begin_provider, Cont_provider ]:
        if type(line) == Begin_provider:
          varlist = []
        buffer = map(strip,line.lower.replace(']','').split(','))
        assert len(buffer) > 1
        v = buffer[1]
        varlist.append(v)
        variable_list = find_variables_in_line(line)
        try:
          variable_list.remove(variables[v].same_as)
        except ValueError:
          print v, variables[v].same_as
          raise
        append( (variable_list,line) )
      elif type(line) == End_provider:
        varlist = []
        append( ([],line) )
      elif type(line) == Provide:
        l = line.lower.split()[1:]
        l = filter(lambda x: x not in varlist, l)
        for v in l:
          if v not in variables:
            error.fail(line,"Variable %s is unknown"%(v))
        append( (l,Provide(line.i,"",line.filename)) )
        append( (l,Simple_line(line.i,"!%s"%(line.text),line.filename)) )
      elif type(line) == NoDep:
        l = line.lower.split()[1:]
        for v in l:
          if v not in variables:
            error.fail(line,"Variable %s is unknown"%(v))
        l = map(lambda x: "-%s"%(x), l)
        append( (l,Simple_line(line.i,"!%s"%(line.text),line.filename)) )
      elif type(line) in [ Touch, SoftTouch ]:
        vars = line.lower.split()
        if len(vars) < 2:
          error.fail(line,"Syntax error")
        vars = vars[1:]
        def fun(x):
          main = variables[x].same_as
          return main
        main_vars = make_single( map(fun, vars) )
        check_touch(line,vars,main_vars)
        txt = " ".join(vars)
        append ( (vars,Simple_line(line.i,"!",line.filename)) )
        append ( ([],Simple_line(line.i,"! >>> TOUCH %s"%(txt,),line.filename)) )
        def fun(x):
          if x not in variables:
            error.fail(line,"Variable %s unknown"%(x,))
          return [ ([],Simple_line(line.i," call touch_%s"%(x,),line.filename)),
                   ([],Use(line.i," use %s"%(variables[x].fmodule), line.filename)) ]
        result += flatten(map( fun, main_vars ))
        def fun(x):
          if x not in variables:
            error.fail(line,"Variable %s unknown"%(x,))
          return ([],Simple_line(line.i," %s_is_built = .True."%(x,),line.filename))
        result += map( fun, main_vars[:-1] )
        if type(line) == SoftTouch:
          append ( ([],Simple_line(line.i,"! <<< END TOUCH (Soft)",line.filename)) )
        else:
          append ( ([],Provide_all(line.i,"! <<< END TOUCH",line.filename)) )
      elif type(line) == Call:
        l = find_variables_in_line(line)
        l = filter(lambda x: x not in varlist, l)
        sub = find_subroutine_in_line(line)
        if sub not in subroutines:
          t = Simple_line
          append( (l,Simple_line(line.i,line.text,line.filename)) )
        else:
          append( (l,line) )
          if subroutines[sub].touches != []:
            append( ([],Provide_all(line.i,"",line.filename)) )
      elif type(line) == Free:
        vars = line.lower.split()
        vars = vars[1:]
        append( ([],Simple_line(line.i,"!%s"%(line.text),line.filename)) )
        use = map(lambda x: "  use %s"%(variables[x].fmodule),vars)
        for var in vars:
          result += map(lambda x: ([],Use(line.i,x,line.filename)),
            make_single(use))
          result += map(lambda x: ([],Simple_line(line.i,x,line.filename)),
            variables[var].free)
      elif type(line) == Irp_read:
        append( ([],Simple_line(line.i,"!%s"%(line.text),line.filename)) )
      elif type(line) == Irp_write:
        append( ([],Simple_line(line.i,"!%s"%(line.text),line.filename)) )
      elif type(line) in [ Begin_doc, End_doc, Doc ]:
        pass
      else:
        l = find_variables_in_line(line)
        l = filter(lambda x: x not in varlist, l)
        append( (l,line) )
    return result

 #main_result = []
 #for filename,text in preprocessed_text:
 #  main_result.append( (filename, func(filename,text)) )
 #return main_result
  return parallel_loop(func,preprocessed_text)

update_variables()
parsed_text = get_parsed_text()


######################################################################

def move_to_top(text,t):
  assert type(text) == list
  assert t in [ NoDep, Declaration, Implicit, Use, Cont_provider ]

  inside = False
  for i in range(len(text)):
    vars, line = text[i]
    if type(line) in [ Begin_provider, Program, Subroutine, Function ]:
      begin = i
      inside = True
    elif type(line) in [ End_provider, End ]:
      inside = False
    elif type(line) == t:
      if inside:
        text.pop(i)
        begin += 1
        text.insert(begin,(vars,line))

  return text

result = []
for filename,text in parsed_text:
  text = move_to_top(text,NoDep)
  text = move_to_top(text,Declaration)
  text = move_to_top(text,Implicit)
  text = move_to_top(text,Use)
  text = move_to_top(text,Cont_provider)
  result.append ( (filename,text) )
parsed_text = result


######################################################################
def build_sub_needs():
  # Needs
  for filename, text in parsed_text:
    sub = None
    in_program = False
    for vars,line in text:
      if type(line) in [ Subroutine, Function ]:
        subname = find_subname(line)
        sub = subroutines[subname]
        sub._needs = []
        sub._to_provide = []
      elif type(line) == End:
        if not in_program:
          sub._needs = make_single(sub._needs)
          sub._to_provide = make_single(sub._to_provide)
          sub = None
      elif type(line) == Program:
        in_program = True
      if sub is not None:
        if type(line) == Declaration:
          sub._to_provide += vars
        sub._needs += vars

build_sub_needs()

#####################################################################

def add_subroutine_needs():
  main_result = []
  for filename, text in parsed_text:
    result = []
    append = result.append
    for vars,line in text:
      if type(line) == Call:
        subname = find_subname(line)
        vars += subroutines[subname].to_provide
      append( (vars,line) )
    main_result.append( (filename, result) )
  return main_result

parsed_text = add_subroutine_needs()

######################################################################
def move_variables():

  def func(filename, text):
    result = []
    append = result.append
    # 1st pass
    varlist = []
    ifvars = []
    elsevars = []
    old_varlist = []
    old_ifvars = []
    old_elsevars = []
    revtext = list(text)
    revtext.reverse()
    try:
      for vars,line in revtext:
        if type(line) in [ End_provider,End ]:
          varlist = []
          append( ([],line) )
        elif type(line) in [ Endif, End_select ]:
          old_ifvars.append( list(ifvars) )
          old_elsevars.append( list(elsevars) )
          old_varlist.append( list(varlist) )
          varlist = []
          append( ([],line) )
        elif type(line) == Else:
          elsevars += list(varlist)
          append( (varlist,line) )
          varlist = []
        elif type(line) in [ Elseif, Case ]:
          ifvars += list(varlist)
          append( (varlist,line) )
          if vars != []:
            varlist = old_varlist.pop()
            varlist += vars
            old_varlist.append( list(varlist) )
          varlist = []
        elif type(line) in [ If, Select ]:
          ifvars += list(varlist)
          append( (varlist,line) )
          vars += filter(lambda x: x in elsevars, ifvars)
          ifvars = old_ifvars.pop()
          elsevars = old_elsevars.pop()
          varlist = old_varlist.pop() + vars
        elif type(line) in [ Begin_provider, Program, Subroutine, Function ]:
          varlist += vars
          append( (varlist,line) )
          if old_varlist != [] \
           or old_ifvars != [] \
           or old_elsevars != []:
            error.fail(line,"End if missing")
          varlist = []
        elif type(line) in (Provide,Provide_all):
          append( (vars,line) )
        else:
          varlist += vars
          append( ([],line) )
    except:
      error.fail(line,"Unable to parse file")

    result.reverse()

    # 2nd pass
    text = result
    result = []
    append = result.append
    old_varlist = []
    varlist = []
    try:
      for vars,line in text:
        if vars != []:
          vars = make_single(vars)
        if type(line) in [ Begin_provider, Program, Subroutine, Function ]:
          varlist = list(vars)
        elif type(line) in [ If, Select ]:
          old_varlist.append(varlist)
          vars = filter(lambda x: x not in varlist,vars)
          varlist = make_single(varlist + vars)
          assert old_varlist is not varlist
        elif type(line) in [ Elseif, Else, Case ]:
          varlist = old_varlist.pop()
          old_varlist.append(varlist)
          vars = filter(lambda x: x not in varlist,vars)
          varlist = make_single(varlist + vars)
          assert old_varlist is not varlist
        elif type(line) in [ Endif, End_select ]:
          varlist = old_varlist.pop()
        elif type(line) == Provide_all:
          vars += varlist
        elif type(line) in [ End_provider, End ]:
          assert old_varlist == []
          varlist = []
        for v in vars[:]:
          if v[0] == '-':
            vars.remove(v)
            vars.remove(v[1:])
        result.append( (vars,line) )
    except:
      error.fail(line,"Unable to parse file")
    return result

  main_result = []
  for filename,text in parsed_text:
    main_result.append( (filename, func(filename,text)) )
  return main_result
 #return parallel_loop(func,parsed_text)

parsed_text = move_variables()

######################################################################
def build_needs():
  # Needs
  for filename, text in parsed_text:
    var = None
    for vars,line in text:
      if type(line) == Begin_provider:
        buffer = map(strip,line.lower.replace(']',',').split(','))
        var = variables[buffer[1]]
        var.needs = []
        var.to_provide = vars
      elif type(line) == End_provider:
        var.needs = make_single(var.needs)
        var.to_provide = make_single(var.to_provide)
        var = None
      if var is not None:
        var.needs += vars
        if type(line) == Call:
          subname = find_subname(line)
          var.needs += subroutines[subname].needs
        elif type(line) in [ \
          Simple_line,  Assert,
          Do         ,  If,
          Elseif     ,  Select,
        ]:
          funcs = find_funcs_in_line(line)
          for f in funcs:
            var.needs += subroutines[f].needs
  for v in variables:
    main = variables[v].same_as
    if main != v:
      variables[v].needs = variables[main].needs
      variables[v].to_provide = variables[main].to_provide

  # Needed_by
  for v in variables:
    variables[v].needed_by = []
  for v in variables:
    main = variables[v].same_as
    if main != v:
      variables[v].needed_by = variables[main].needed_by
  for v in variables:
    var = variables[v]
    if var.is_main:
      for x in var.needs:
        variables[x].needed_by.append(var.same_as)
  for v in variables:
    var = variables[v]
    var.needed_by = make_single(var.needed_by)


build_needs()

result = []
for filename,text in parsed_text:
  text = move_to_top(text,NoDep)
  text = move_to_top(text,Declaration)
  text = move_to_top(text,Implicit)
  text = move_to_top(text,Use)
  text = move_to_top(text,Cont_provider)
  result.append ( (filename,text) )
parsed_text = result


######################################################################
from command_line import command_line

def check_opt():
  if not command_line.do_checkopt:
    return

  for filename, text in parsed_text:
    do_level = 0
    for vars,line in text:
     if not type(line) == Provide_all:
      if do_level > 0 and vars != []:
        print "Optimization: %s line %d"%(line.filename,line.i)
        for v in vars:
          print "  PROVIDE ",v
      if type(line) == Do:
        do_level += 1
      elif type(line) == Enddo:
        do_level -= 1
check_opt()

######################################################################
def perform_loop_substitutions():
  main_result = []
  for filename, text in parsed_text:
    result = []
    append = result.append
    for vars,line in text:
      if type(line) in [ Do, If, Elseif ] :
        for k,v in command_line.substituted.items():
          reg = v[1]
          while reg.search(line.text) is not None:
            line.text = re.sub(reg,r'\1%s\3', line.text,count=1)%v[0]
      append( (vars,line) )
    main_result.append( (filename, result) )
  return main_result

parsed_text = perform_loop_substitutions()

######################################################################
if __name__ == '__main__':
 for i in range(len(parsed_text)):
  if parsed_text[i][0] == sys.argv[1]:
   print '!-------- %s -----------'%(parsed_text[i][0])
   for line in parsed_text[i][1]:
     print line[1]
     print line[0], line[1].filename
#for i in subroutines:
#  print i, subroutines[i].needs, subroutines[i].to_provide



import sys

import setuptools

with open("../README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'irpf90',
  version = '1.7.5',
  scripts = ["irpf90", "irpman", "irpf90_indent"],
  author = 'Anthony Scemama',
  author_email = 'scemama@irsamc.ups-tlse.fr',
  description = 'IRPF90 is a Fortran90 preprocessor written in Python for programming using the Implicit Reference to Parameters (IRP) method. It simplifies the development of large fortran codes in the field of scientific high performance computing.',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://irpf90.ups-tlse.fr",
  download_url = 'https://gitlab.com/scemama/irpf90/archive/v1.7.4.tar.gz',
  packages=setuptools.find_packages(),
  classifiers=[
         "Programming Language :: Python :: 2",
         "Programming Language :: Fortran",
         "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
         "Operating System :: POSIX :: Linux",
     ],
  keywords = ['programming', 'fortran', 'IRP'], 
 )



#!/usr/bin/env python2
#   IRPF90 is a Fortran90 preprocessor written in Python for programming using
#   the Implicit Reference to Parameters (IRP) method.
#   Copyright (C) 2009 Anthony SCEMAMA 
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#   Anthony Scemama
#   LCPQ - IRSAMC - CNRS
#   Universite Paul Sabatier
#   118, route de Narbonne      
#   31062 Toulouse Cedex 4      
#   scemama@irsamc.ups-tlse.fr


import os

import util
import makefile
import irpf90_t
from command_line import command_line

initialized = False


def init():

  global initialized
  if initialized:
     return

  # Create directories
  for dir in [ irpf90_t.irpdir, irpf90_t.mandir ]:
    try:
      wd = os.getcwd()
      os.chdir(dir)
      os.chdir(wd)
    except OSError:
      os.mkdir(dir)

  for dir in command_line.include_dir:
    dir = irpf90_t.irpdir+dir
    try:
      wd = os.getcwd()
      os.chdir(dir)
      os.chdir(wd)
    except OSError:
      os.mkdir(dir)

  # Create makefile
  makefile.create()
  
  # Copy current files in the irpdir
  for dir in ['./']+command_line.include_dir:
    try:
      os.stat(dir)
    except:
      print dir,'not in dir'
      continue
    for filename in os.listdir(dir):
      filename = dir+filename
      if not filename.startswith(".") and not os.path.isdir(filename):
        try:
          file  = open(filename,"r")
        except IOError:
          if command_line.do_warnings:
              print "Warning : Unable to read file %s."%(filename)
        else:
          buffer = file.read()
          file.close()
          if not util.same_file(irpf90_t.irpdir+filename,buffer):
            file = open(irpf90_t.irpdir+filename,"w")
            file.write(buffer)
            file.close()

  initialized = True

